-----------------------------------------------------------------------------

local AceOO = AceLibrary("AceOO-2.0")
local SML = AceLibrary("SharedMedia-1.0")
local Dewdrop = AceLibrary("Dewdrop-2.0")
local AssessmentBarInterface = Assessment:NewModule("BarInterface", "AceEvent-2.0")
AssessmentBarInterface.db = Assessment:AcquireDBNamespace("BarInterface")
Assessment:RegisterDefaults("BarInterface", "profile", {
	windows =
	{
-- Set on first login if it doesn't exist, and defaultWindowProfile copied in
--		count = 1,
	},
})

local AssessmentDisplaySets = Assessment:GetModule("DisplaySets")

-----------------------------------------------------------------------------

local SML = AceLibrary("SharedMedia-1.0")
local SML_statusbar = SML:List("statusbar")
local SML_font = SML:List("font")

local Dewdrop = AceLibrary:HasInstance("Dewdrop-2.0") and AceLibrary("Dewdrop-2.0")

-----------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("AssessmentBarInterface")
L:RegisterTranslations("enUS", function() return {
	["Assessment"] = true,

	["Windows"] = true,
	["Window options."] = true,
	["Display Sets"] = true,
	["Display set options."] = true,
	["Add New Window"] = true,
	["Add a new assessment window to the user interface."] = true,

	["Window %d"] = true,
	["Options for assessment window %d"] = true,
	["Display Set"] = true,
	["Choose what to show for this window"] = true,
	["Visible"] = true,
	["Make the assessment window visible"] = true,

	["Lock"] = true,
	["Enable to prevent this window from being draggable."] = true,

	["Update Rate"] = true,
	["Controls the maximum number of times per second to update this window. Use 0 for continuous updates."] = true,

	["Backdrop Color"] = true,
	["Change the backdrop color and alpha for the window."] = true,
	["Highlight Color"] = true,
	["Change the highlight color and alpha for the window."] = true,
	["Border Style"] = true,
	["Change the border style for the window."] = true,
	["None"] = true,
	["Default"] = true,
	["Textured"] = true,
	["Square"] = true,
	["Border Color"] = true,
	["Change the border color and alpha for the window."] = true,

	["Bar Options"] = true,
	["Bar options for this window"] = true,
	["Bar Count"] = true,
	["Number of bars to show"] = true,
	["Texture"] = true, 
	["Sets the bar texture"] = true,

	["Title Options"] = true,
	["Title options."] = true,
	["Enable"] = true,
	["Enable title text."] = true,
	["Alignment"] = true,
	["Select how to align the title."] = true,
	["Left"] = true,
	["Center"] = true,
	["Right"] = true,
	["Font"] = true,
	["Font for title."] = true,
	["Font Size"] = true,
	["Font size for title."] = true,
	["Font Shadow"] = true,
	["Show font shadow on title text"] = true,

	["Outline"] = true,
	["Sets the outline for title text"] = true,
	["None"] = true,
	["Normal"] = true,
	["Thick"] = true,

	["Title Height"] = true,
	["Title height."] = true,
	["Background Color"] = true,
	["Change the title backdrop color."] = true,
	["Text Color"] = true,
	["Change the title text color."] = true,

	["This display set cannot be removed currently because one or more windows require its data."] = true,

	["Remove"] = true,
	["Remove this window from display"] = true,
} end)

-----------------------------------------------------------------------------

local defaultWindowProfile =
{
	displaySetStack		= { "Assessment" },
	displaySetStackLevel= 1,
	updateRate			= 10,
	title = 
	{
		enable			= true,
		alignment		= "CENTER",
		font			= "Arial Narrow",
		fontSize		= 9,
		outline			= "OUTLINE",
		shadow			= true,
		height			= 10,
		textColor		= { 1, 1, 1, 1 },
		backgroundColor	= { 0, 0, 0, 1 }
	},
	bar = 
	{ 
		count			= 10,
		texture			= "Blizzard",
		highlightColor	= { 1, 1, 0, 0.35 },
	},
	frame = 
	{ 
		lock			= false,
		width			= 110,
		height			= 140, 
		point			= "CENTER",
		relativePoint	= "CENTER",
		x				= 0, 
		y				= 0, 
		borderStyle		= "Default",
		borderColor		= { 1, 1, 1, 0.75 },
		backdropColor	= { 0, 0, 0, 0.75 },
		visible			= true,
	},
}

-----------------------------------------------------------------------------

Assessment.Options.args.windows =
{
	type = "group",
	name = L["Windows"],
	desc = L["Window options."],
	order = 60,
	args = { },
}

local windowsArgs =
{ 
	add =
	{
		type = "execute",
		name = L["Add New Window"],
		desc = L["Add a new assessment window to the user interface."],
		func = function() AssessmentBarInterface:AddWindow() end,
		order = -1
	},
}

-----------------------------------------------------------------------------

local get = AssessmentBarInterface.GetOptionValue
local set = AssessmentBarInterface.SetOptionValue

local function windowGet(passValue)
	return get(AssessmentBarInterface.db.profile.windows[passValue[1]], passValue[2])
end

local function windowSet(passValue, value)
	set(AssessmentBarInterface.db.profile.windows[passValue[1]], passValue[2], value)
	AssessmentBarInterface:UpdateWindowOptions(passValue[1])
end

local function windowGetDisplaySet(passValue)
	local windowOptions = AssessmentBarInterface.db.profile.windows[passValue[1]]
	return ("$"):split(windowOptions.displaySetStack[windowOptions.displaySetStackLevel], 2)
end

local function windowSetDisplaySet(passValue, value)
	local windowOptions = AssessmentBarInterface.db.profile.windows[passValue[1]]
	
	windowOptions.displaySetStackLevel = 1
	windowOptions.displaySetStack[1] = value

	AssessmentBarInterface:UpdateWindowOptions(passValue[1])
end

local function windowRemove(passValue)
	AssessmentBarInterface:RemoveWindow(passValue[1])
end

-----------------------------------------------------------------------------

local displaySetValidate		= { }

-----------------------------------------------------------------------------

local windowOptionsTemplate =
{
	type = "group",
--	name = ["Window %d"] = true,
--	desc = ["Options for assessment window %d"] = true,
	args =
	{
		visible =
		{
			type = "toggle",
			name = L["Visible"],
			desc = L["Make the assessment window visible"],
			get = windowGet,
			set = windowSet,
			passValue = "frame.visible",
			order = 1,
		},
		displaySet = 
		{
			type = "text",
			name = L["Display Set"],
			desc = L["Choose what to show for this window"],
			get = windowGetDisplaySet,
			set = windowSetDisplaySet,
			passValue = "displaySet",
			validate = displaySetValidate,
			order = 2,
		},
		lock =
		{
			type = "toggle",
			name = L["Lock"],
			desc = L["Enable to prevent this window from being draggable."],
			get = windowGet,
			set = windowSet,
			passValue = "frame.lock",
		},
		backdropColor = 
		{
			type = "color",
			name = L["Backdrop Color"],
			desc = L["Change the backdrop color and alpha for the window."],
			get = function(passValue) return unpack(windowGet(passValue)) end,
			set = function(passValue, r, g, b, a) windowSet(passValue, { r, g, b, a }) end,
			passValue = "frame.backdropColor",
			hasAlpha = true,
		},
		borderStyle = 
		{
			type = "text",
			name = L["Border Style"],
			desc = L["Change the border style for the window."],
			get = windowGet,
			set = windowSet,
			validate = { None = L["None"], Default = L["Default"], Textured = L["Textured"], Square = L["Square"] },
			passValue = "frame.borderStyle",
		},
		borderColor = 
		{
			type = "color",
			name = L["Border Color"],
			desc = L["Change the border color and alpha for the window."],
			get = function(passValue) return unpack(windowGet(passValue)) end,
			set = function(passValue, r, g, b, a) windowSet(passValue, { r, g, b, a }) end,
			passValue = "frame.borderColor",
			hasAlpha = true,
		},
		updateRate =
		{
			type = "range",
			name = L["Update Rate"],
			desc = L["Controls the maximum number of times per second to update this window. Use 0 for continuous updates."],
			min  = 0,
			max  = 60,
			step = 1,
			passValue = "updateRate",
			get = windowGet,
			set = windowSet,
		},
		bars =
		{
			type = "group",
			name = L["Bar Options"],
			desc = L["Bar options for this window"],
			args =
			{
				count =
				{
					type = "range",
					name = L["Bar Count"],
					desc = L["Number of bars to show"],
					min	 = 1,
					max  = 50,
					step = 1,
					get = windowGet,
					set = windowSet,
					passValue = "bar.count",
					order = 1,
				},
				highlightColor = 
				{
					type = "color",
					name = L["Highlight Color"],
					desc = L["Change the highlight color and alpha for the window."],
					get = function(passValue) return unpack(windowGet(passValue)) end,
					set = function(passValue, r, g, b, a) windowSet(passValue, { r, g, b, a }) end,
					passValue = "bar.highlightColor",
					hasAlpha = true,
				},
				texture =
				{
					type = "text", 
					name = L["Texture"], 
					desc = L["Sets the bar texture"],
					get = windowGet,
					set = windowSet,
					validate = SML_statusbar,
					passValue = "bar.texture",
					order = 2,
				},
			},
		},
		title =
		{
			type = "group",
			name = L["Title Options"],
			desc = L["Title options."],
			args =
			{
				enable = 
				{
					type = "toggle",
					name = L["Enable"],
					desc = L["Enable title text."],
					get = windowGet,
					set = windowSet,
					passValue = "title.enable",
					order = 1,
				},
				alignment =
				{
					type = "text",
					name = L["Alignment"],
					desc = L["Select how to align the title."],
					get = windowGet,
					set = windowSet,
					passValue = "title.alignment",
					validate = { LEFT = L["Left"], CENTER = L["Center"], RIGHT = L["Right"] },
				},
				font =
				{
					type = "text",
					name = L["Font"],
					desc = L["Font for title."],
					get = windowGet,
					set = windowSet,
					passValue = "title.font",
					validate = SML_font,
				},
				height =
				{
					type = "range",
					name = L["Title Height"],
					desc = L["Title height."],
					min  = 0,
					max  = 32,
					step = 1,
					get = windowGet,
					set = windowSet,
					passValue = "title.height",
				},
				fontSize =
				{
					type = "range",
					name = L["Font Size"],
					desc = L["Font size for title."],
					min = 4,
					max = 32,
					step = 1,
					get = windowGet,
					set = windowSet,
					passValue = "title.fontSize",
				},
				shadow =
				{
					type = "toggle",
					name = L["Font Shadow"],
					desc = L["Show font shadow on title text"],
					get = windowGet,
					set = windowSet,
					passValue = "title.shadow",
				},
				backgroundColor = 
				{
					type = "color",
					name = L["Background Color"],
					desc = L["Change the title backdrop color."],
					get = function(passValue) return unpack(windowGet(passValue)) end,
					set = function(passValue, r, g, b, a) windowSet(passValue, { r, g, b, a }) end,
					passValue = "title.backgroundColor",
					hasAlpha = true,
				},
				textColor = 
				{
					type = "color",
					name = L["Text Color"],
					desc = L["Change the title text color."],
					get = function(passValue) return unpack(windowGet(passValue)) end,
					set = function(passValue, r, g, b, a) windowSet(passValue, { r, g, b, a }) end,
					passValue = "title.textColor",
					hasAlpha = true,
				},
				outline =
				{
					type = "text",
					name = L["Outline"],
					desc = L["Sets the outline for title text"],
					get = windowGet,
					set = windowSet,
					passValue = "title.outline",
					validate = { [""] = L["None"], OUTLINE = L["Normal"], THICKOUTLINE = L["Thick"] },
				},
			},
		},
		remove =
		{
			type = "execute",
			name = L["Remove"],
			desc = L["Remove this window from display"],
			func = windowRemove,
			passValue = false,
			order = -1,
		},
	}
}

-----------------------------------------------------------------------------

-- This thing is created on demand, and reparented as necessary
local uiBarHighlightFrame

local UIBar = AceOO.Class()

local function OnUIBarEnter(self, motion)
	local frame = self.uiBar.frame
	if not uiBarHighlightFrame then
		uiBarHighlightFrame = frame:CreateTexture(nil, "HIGHLIGHT")
		uiBarHighlightFrame:SetTexture("Interface\\QuestFrame\\UI-QuestTitleHighlight")
		uiBarHighlightFrame:SetBlendMode("ADD")
	end
	uiBarHighlightFrame:SetVertexColor(unpack(self.uiBar.windowOptions.bar.highlightColor))
	uiBarHighlightFrame:SetParent(frame)
	uiBarHighlightFrame:ClearAllPoints()
	uiBarHighlightFrame:SetAllPoints(frame)
	uiBarHighlightFrame:Show()
end

local function OnUIBarLeave(self, motion)
	uiBarHighlightFrame:Hide()
end

local function OnUIBarMouseDown(self, button)
	if button == "LeftButton" then
		self:GetParent().window:OnBarClicked(self.uiBar)
	elseif button == "RightButton" then
		self:GetParent().window:PopDisplaySetStack()
	end
end

function UIBar.prototype:init(parent, windowOptions, displaySet, index)
	UIBar.super.prototype.init(self)

	local frame = CreateFrame("Frame", nil, parent)
	frame.uiBar = self
	self.frame = frame

	local bar = frame:CreateTexture()
	self.bar = bar
	self.texts = { }
	self.barColorMethods = { }

	bar:SetPoint("TOPLEFT", frame, "TOPLEFT")
	bar:SetPoint("BOTTOMLEFT", frame, "BOTTOMLEFT")
	
	frame:SetScript("OnEnter", OnUIBarEnter)
	frame:SetScript("OnLeave", OnUIBarLeave)
	frame:SetScript("OnMouseDown", OnUIBarMouseDown)
	frame:EnableMouse()

	self:SetOptions(parent, windowOptions, displaySet, index)
end

function UIBar.prototype:SetOptions(parent, windowOptions, displaySet, index)
	self.windowOptions = windowOptions
	self.displaySet = displaySet

	local bar = self.bar

	for i in ipairs(self.barColorMethods) do
		self.barColorMethods[i] = nil
	end
	for i = 1, displaySet.colorTypeList.count do
		self.barColorMethods[i] = Assessment.ColorType[displaySet.colorTypeList[i]].method
	end
	self.barColorMethods[#self.barColorMethods+1] = function() return displaySet.defaultColor end

	bar:SetTexture(SML:Fetch("statusbar", windowOptions.bar.texture))
	bar:SetVertexColor(1, 1, 1, 1)

	for i = 1, self.displaySet.text.count do
		if not self.texts[i] then
			self.texts[i] = self.frame:CreateFontString()
		end
		self.texts[i]:Show()
	end

	for i = self.displaySet.text.count + 1, #self.texts do
		self.texts[i]:Hide()
	end
	
	self:PlaceBar(parent, index)
end

function UIBar.prototype:PlaceBar(parent, index)
	local frame = self.frame
	local bar = self.bar
	local verticalOffset = -5 - (self.windowOptions.title.enable and self.windowOptions.title.height or 0)
	local height = (parent:GetHeight()+verticalOffset-5) / self.windowOptions.bar.count
	
	self.maximumBarWidth = (self.windowOptions.frame.width-10)
	
	frame:ClearAllPoints()
	frame:SetPoint("TOPLEFT", parent, "TOPLEFT", 5, -(index-1)*height+verticalOffset)
	frame:SetPoint("BOTTOMRIGHT", parent, "TOPRIGHT", -5, -index*height+verticalOffset)
	
	for i = 1, self.displaySet.text.count do
		local text = self.texts[i]
		local options = self.displaySet.text[i]
		
		text:ClearAllPoints()
		text:SetPoint("LEFT", frame, "LEFT", options.offsets.left)
		text:SetPoint("RIGHT", frame, "RIGHT", options.offsets.right)
		text:SetPoint("TOP", frame, "TOP", 0, options.offsets.vertical)
		text:SetPoint("BOTTOM", frame, "BOTTOM", 0, options.offsets.vertical)

		text:SetAlpha(options.alpha)

		text:SetFont(SML:Fetch("font", options.font), options.fontSize, options.outline)
		if options.shadow then
			text:SetShadowOffset(1, -1)
		else
			text:SetShadowOffset(0, 0)
		end
		text:SetJustifyH(options.alignment)
	end
end

function UIBar.prototype:SetFraction(fraction)
	local width = fraction * self.maximumBarWidth
	local bar = self.bar
	if width == 0 then
		bar:Hide()
	else
		bar:SetWidth(width)
		bar:Show()
	end
end

function UIBar.prototype:Hide()
	self.frame:Hide()
end

function UIBar.prototype:Update(data, barNormalizeMultiplier, textMethods)
	-- Update Bar
	self:SetFraction(data.barValue * barNormalizeMultiplier)
	
	local color
	for i,f in ipairs(self.barColorMethods) do
		color = f(data.fightData, data.key)
		if color then break end
	end
	self.bar:SetVertexColor(unpack(color))
	
	-- Update Texts
	for i = 1, self.displaySet.text.count do
		local text = self.texts[i]
		text:SetText(textMethods[i](data))
	end
	self.frame:Show()
end

-----------------------------------------------------------------------------

local borderStyles =
{
	None		= { edgeFile = "", inset = 5 },
	Default		= { edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", inset = 5 },
	Textured	= { edgeFile = "", inset = 3 },
	Square		= { edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border", inset = 5 },
}

local UIWindow = AceOO.Class("AceEvent-2.0")

function UIWindow.prototype:init(windowOptions, displaySet)
	UIWindow.super.prototype.init(self)
	local frame = CreateFrame("Frame", nil, UIParent)
	frame.window = self

	self.dependentData		= { }
	self.dataUpdateTimes	= { }
	self.textMethods		= { }
	self.listTypeParameters = { }

	self.frame			= frame
	self.bars			= { }
	
	self.titleText = frame:CreateFontString()
	self.titleTexture = frame:CreateTexture()

	self:SetWindowOptions(windowOptions)
end

local bgTable = 
{
    bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
    tile = true,
    tileSize = 16,
    edgeSize = 16,
	insets = { left = inset, right = inset, top = inset, bottom = inset }
}

local isWindowSizing, wasWindowDragged, wasWindowSizeChanged

local function OnUIWindowMouseDown(this, button)
	if button == "LeftButton" then
		if not this.window.windowOptions.frame.lock then 
			local mouseDownX, mouseDownY = GetCursorPosition(UIParent);

			local effectiveScale = this:GetEffectiveScale();
			local left = this:GetLeft() * effectiveScale;
			local top = this:GetTop() * effectiveScale;
			local right = this:GetRight() * effectiveScale;
			local bottom = this:GetBottom() * effectiveScale;
			
			local edge = "";
			if (abs(mouseDownY - bottom) < 7) then
				edge = "BOTTOM";
			elseif (abs(mouseDownY - top) < 7) then
				edge = "TOP"
			end
			
			if (abs(mouseDownX - left) < 7) then
				edge = edge.."LEFT";
			elseif (abs(mouseDownX - right) < 7) then
				edge = edge.."RIGHT";
			end
			
			if edge ~= "" then
				this:StartSizing(edge)
				isWindowSizing = true
				wasWindowSizeChanged = nil
			end
		end
	elseif button == "RightButton" then
		this.window:PopDisplaySetStack()
	end
end

local function OnUIWindowMouseUp(this, button)
	if button ~= "LeftButton" then return end
	
	if isWindowSizing then
		this:StopMovingOrSizing() 
		isWindowSizing = nil
		if wasWindowSizeChanged then
			wasWindowSizeChanged = nil
		else
			this.window:OnClick(button)
		end
	elseif wasWindowDragged then
		wasWindowDragged = nil
	else
		this.window:OnClick(button)
	end
end

local function OnUIWindowDragStart(this)
	wasWindowDragged = true
	if not this.window.windowOptions.frame.lock then 
		this:StartMoving() 
	end
end

local function OnUIWindowDragStop(this)
	this:StopMovingOrSizing() 
	local point,_,relativePoint,x,y = this:GetPoint(1)
	local frameOptions = this.window.windowOptions.frame
	frameOptions.point = point
	frameOptions.relativePoint = relativePoint
	frameOptions.x = x
	frameOptions.y = y
end

local function OnUIWindowSizeChanged(this, width, height)
	wasWindowSizeChanged = true
	local window = this.window
	local frameOptions = window.windowOptions.frame
	frameOptions.width = width 
	frameOptions.height = height

	local point,_,relativePoint,x,y = this:GetPoint(1)
	frameOptions.point = point
	frameOptions.relativePoint = relativePoint
	frameOptions.x = x
	frameOptions.y = y

	window:LayoutBars()
	window.forceUpdate = true
	window:Update()
end

function UIWindow.prototype:SetWindowOptions(windowOptions)
	local displaySet, displaySetParameterText = AssessmentBarInterface:GetDisplaySet(windowOptions)

	-- Clear some tables
	for k in pairs(self.dependentData) do self.dependentData[k] = nil end
	for k in pairs(self.dataUpdateTimes) do self.dataUpdateTimes[k] = nil end
	for k in ipairs(self.textMethods) do self.textMethods[k] = nil end

	self.topBarOffset	= 0
	self.forceUpdate	= true
	self.displaySet		= displaySet
	self.windowOptions	= windowOptions

	self.listTypeMethod = Assessment.ListType[displaySet.listType].method
	Assessment:ReleaseTable(self.listTypeParameters)
	self.listTypeParameters = displaySetParameterText and { (","):split(displaySetParameterText) } or Assessment:NewTable()
	
	local compiledFilterValue = Assessment:CompileTag(displaySet.filterTag, "raw")
	self:MergeRequiredData(compiledFilterValue.data)
	self.filterMethod = compiledFilterValue.method

	for i=1,displaySet.colorTypeList.count do
		self:MergeRequiredData(Assessment.ColorType[displaySet.colorTypeList[i]].data)
	end
	self:MergeRequiredData(Assessment.ListType[displaySet.listType].data)

	if displaySet.descendingSort then
		self.sortDirectionMethod = function(a, b) if a.sortValue == b.sortValue then
													return a.key < b.key	-- always sort keys ascending if the sort values are equal
												  else
												    return a.sortValue > b.sortValue
												  end
								   end
	else
		self.sortDirectionMethod = function(a, b) if a.sortValue == b.sortValue then
													return a.key < b.key
												  else
												    return a.sortValue < b.sortValue
												  end
								   end
	end

	local compiledBarValue = Assessment:CompileTag(displaySet.barValueTag, "number")
	self.barValueMethod = compiledBarValue.method
	self:MergeRequiredData(compiledBarValue.data)
	
	for i=1,displaySet.text.count do
		local compiledTextValue = Assessment:CompileTag(displaySet.text[i].tag, "string")
		self.textMethods[i] = compiledTextValue.method
		self:MergeRequiredData(compiledTextValue.data)
	end

	local compiledSortValue = Assessment:CompileTag(displaySet.sortKeyTag, "raw")
	self:MergeRequiredData(compiledSortValue.data)
	self.sortMethod = compiledSortValue.method

	-------

	local frame = self.frame
	local frameOptions = windowOptions.frame
	frame:ClearAllPoints()
	frame:SetPoint(frameOptions.point, UIParent, frameOptions.relativePoint, frameOptions.x, frameOptions.y)
	frame:SetWidth(frameOptions.width)
	frame:SetHeight(frameOptions.height)
	frame:SetMovable(true)
	frame:SetResizable(true)
	frame:SetMinResize(32,32)
	frame:EnableMouse(true)
	frame:EnableMouseWheel(true)
	frame:RegisterForDrag("LeftButton")

	frame:SetScript("OnMouseWheel", function(this, delta) 
										local window = this.window
										if delta < 0 then
											window.topBarOffset = window.topBarOffset + 1
										else
											window.topBarOffset = window.topBarOffset - 1
											if window.topBarOffset < 0 then
												window.topBarOffset = 0
											end
										end
										window.forceUpdate = true
										window:Update()
									end)
	
	frame:SetScript("OnMouseDown",	OnUIWindowMouseDown)
	frame:SetScript("OnMouseUp",	OnUIWindowMouseUp)
	frame:SetScript("OnDragStart",	OnUIWindowDragStart)
	frame:SetScript("OnDragStop",	OnUIWindowDragStop)
	frame:SetScript("OnSizeChanged",OnUIWindowSizeChanged)

	local borderStyle = borderStyles[windowOptions.frame.borderStyle]
	bgTable.edgeFile = borderStyle.edgeFile
	local inset = borderStyle.inset
	local insets = bgTable.insets
	insets.left = inset
	insets.top = inset
	insets.right = inset
	insets.bottom = inset
    frame:SetBackdrop(bgTable)
	frame:SetBackdropBorderColor(unpack(windowOptions.frame.borderColor))
	frame:SetBackdropColor(unpack(windowOptions.frame.backdropColor))

	--------

	local titleText = self.titleText
	local titleOptions = windowOptions.title
	
	if titleOptions.enable then
		titleText:ClearAllPoints()
		titleText:SetPoint("TOPLEFT", frame, "TOPLEFT", 5, -5)
		titleText:SetPoint("BOTTOMRIGHT", frame, "TOPRIGHT", -5, -titleOptions.height-5)
		titleText:SetTextColor(unpack(titleOptions.textColor))

		titleText:SetFont(SML:Fetch("font", titleOptions.font), titleOptions.fontSize, titleOptions.outline)
		if titleOptions.shadow then
			titleText:SetShadowOffset(1, -1)
		else
			titleText:SetShadowOffset(0, 0)
		end
		titleText:SetJustifyH(titleOptions.alignment)

		local compiledTitleTextValue = Assessment:CompileTag(displaySet.titleTag, "string")
		self.titleTextMethod = compiledTitleTextValue.method
		self:MergeRequiredData(compiledTitleTextValue.data)

		titleText:Show()

		local titleTexture = self.titleTexture
		titleTexture:ClearAllPoints()
		titleTexture:SetPoint("TOPLEFT", frame, "TOPLEFT", 5, -5)
		titleTexture:SetPoint("BOTTOMRIGHT", frame, "TOPRIGHT", -5, -titleOptions.height-5)
		titleTexture:SetTexture(unpack(titleOptions.backgroundColor))
		titleTexture:Show()
	else
		titleText:Hide()
		self.titleTexture:Hide()
	end

	--------

	for i = 1, windowOptions.bar.count do
		if self.bars[i] then
			self.bars[i]:SetOptions(frame, windowOptions, displaySet, i)
		else
			self.bars[i] = UIBar:new(frame, windowOptions, displaySet, i)
		end
	end
	for i = windowOptions.bar.count+1, #self.bars do
		self.bars[i]:Hide()
	end

	if frameOptions.visible then
		self:Show()
	else
		self:Hide()
	end
end

function UIWindow.prototype:GetBarIndex(bar)
	for i = 1, self.windowOptions.bar.count do
		if self.bars[i] == bar then
			return i
		end
	end
	error("Internal error: Bar not found")
end

function UIWindow.prototype:PopDisplaySetStack(windowOptions)
	AssessmentBarInterface:PopWindowDisplaySet(self)
end

function UIWindow.prototype:OnClick(buttonName)
	Dewdrop:Open("",
					'children', function() self:UpdateDewdrop() end,
					'cursorX', true,
					'cursorY', true
					)
end

function UIWindow.prototype:UpdateDewdrop()
	local currentDisplaySetName = AssessmentBarInterface:GetDisplaySetName(self.windowOptions)
	local sortList = Assessment:NewTable()
	
	for k,v in AssessmentDisplaySets:IterateDisplaySets() do
		if k == currentDisplaySetName or (v.favorite and v.selectable) then
			table.insert(sortList, k)
		end
	end
	table.sort(sortList)

	for i,k in ipairs(sortList) do
		Dewdrop:AddLine('text', k,
					'checked', k == currentDisplaySetName,
					'func', k ~= currentDisplaySetName and "SetDisplaySet" or nil,
					'arg1', AssessmentBarInterface,
					'arg2', self,
					'arg3', k)
	end

	Assessment:ReleaseTable(sortList)
end

function UIWindow.prototype:Show()
	if self.updater then
		self:CancelScheduledEvent(self.updater)
	end

	local windowOptions = self.windowOptions
	local updateRate = windowOptions.updateRate ~= 0 and 1/windowOptions.updateRate or 0

	self.updater = self:ScheduleRepeatingEvent(self.Update, updateRate, self)
	self.frame:Show()
	self:Update()
end

function UIWindow.prototype:Hide()
	if self.updater then
		self:CancelScheduledEvent(self.updater)
		self.updater = nil
	end
	self.frame:Hide()
end

function UIWindow.prototype:LayoutBars(fightData)
	for i=1, self.windowOptions.bar.count do
		self.bars[i]:PlaceBar(self.frame, i)
	end
end

function UIWindow.prototype:MergeRequiredData(data)
	if not data then return end
	if type(data) == "string" then
		self.dependentData[data] = true
	elseif type(data) == "table" then
		for k in pairs(data) do
			self.dependentData[k] = true
		end
	end
end

function UIWindow.prototype:NeedsUpdate(fightData)
	if self.dependentData.ALWAYSUPDATE then
		return true
	end

	local dataUpdateTimes = self.dataUpdateTimes
	if not fightData then
		if next(dataUpdateTimes) then
			for k in pairs(dataUpdateTimes) do
				dataUpdateTimes[k] = nil
			end
			return true
		end
		if self.forceUpdate then
			self.forceUpdate = nil
			return true
		end
		return false
	end

	local result = false
	for k in pairs(self.dependentData) do
		if dataUpdateTimes[k] ~= fightData.dataUpdateTimes[k] then
			result = true
			dataUpdateTimes[k] = fightData.dataUpdateTimes[k]
		end
	end
	if self.forceUpdate then
		self.forceUpdate = nil
		return true
	end
	return result
end

local sortList = { }
local maximumBarValue
local totalBarValue

function UIWindow.prototype:UpdateSortList(fightData)
	maximumBarValue = 0
	totalBarValue = 0
	for i,v in ipairs(sortList) do
		sortList[i] = Assessment:ReleaseTable(v)
	end

	if fightData then
		for k,v in self.listTypeMethod(fightData, unpack(self.listTypeParameters)) do
			local data = Assessment:NewTable()
			data.key = k
			data.value = v
			data.fightData = fightData
			data.window = self
			
			local barValue = self.barValueMethod(data)
			data.barValue = barValue
			
			if self.filterMethod(data) then
				if barValue > maximumBarValue then
					maximumBarValue = barValue
				end
				totalBarValue = totalBarValue + barValue

				data.barValue = barValue
				sortList[#sortList+1] = data
			else
				Assessment:ReleaseTable(data)
			end			
		end
		
		for _,data in ipairs(sortList) do
			data.totalBarValue = totalBarValue
			data.maximumBarValue = maximumBarValue
			data.sortValue = self.sortMethod(data)
		end

		table.sort(sortList, self.sortDirectionMethod)
	end
end

Assessment.TagData.windowdisplaysetparameter =
{
	method = function(data, value, index)
				return value and data.window.listTypeParameters[index]
			 end,
	parameterTypes = { "number" }
}

local titleData = { key = UnitName("player"), rank = 0 }
function UIWindow.prototype:DoUpdate(fightData)
	self:UpdateSortList(fightData)

	local barNormalizeMultiplier = maximumBarValue ~= 0 and 1 / maximumBarValue or 0

	local topBarOffset = self.topBarOffset
	local barCount = self.windowOptions.bar.count
	if topBarOffset + barCount > #sortList then
		topBarOffset = #sortList - barCount
		if topBarOffset < 0 then
			topBarOffset = 0
		end
		self.topBarOffset = topBarOffset
	end

	local endDisplay = barCount
	if endDisplay > #sortList-topBarOffset then 
		endDisplay = #sortList-topBarOffset 
		for i = endDisplay+1, barCount do
			self.bars[i]:Hide()
		end
	end
	for i = 1, endDisplay do
		local rank = i+topBarOffset
		local data = sortList[rank]
		data.rank = rank
		self.bars[i]:Update(data, barNormalizeMultiplier, self.textMethods)
	end	

	if self.windowOptions.title.enable then
		titleData.fightData = fightData
		titleData.totalBarValue = totalBarValue
		titleData.maximumBarValue = maximumBarValue
		titleData.window = self

		self.titleText:SetText( self.titleTextMethod(titleData) )
	end
end

function UIWindow.prototype:Update()
	local fightData = Assessment:GetFightData(self.displaySet.fightIndex)
	if not self:NeedsUpdate(fightData) then return end
	self:DoUpdate(fightData)
end

function UIWindow.prototype:OnBarClicked(bar)
	if self.displaySet.onClickType == "PUSH_DISPLAY_SET" and self.displaySet.onClickDisplaySet and self.displaySet.onClickData then
		local barIndex = self:GetBarIndex(bar)

		self:UpdateSortList(Assessment:GetFightData(self.displaySet.fightIndex))
		local data = sortList[barIndex+self.topBarOffset]
		
		if data then
			local displaySet = Assessment:CompileTag(self.displaySet.onClickDisplaySet, "string").method(data)
			local parameters = Assessment:CompileTag(self.displaySet.onClickData, "string").method(data)
			AssessmentBarInterface:PushWindowDisplaySet(self, displaySet.."$"..parameters)
		end
	end
end

local function StripColorCodes(s)
	return s:gsub("|c%x%x%x%x%x%x%x%x", ""):gsub("|r", "")
end


function UIWindow.prototype:Report(chatType, limit, channel)
	self:UpdateSortList(Assessment:GetFightData(self.displaySet.fightIndex))
	
	if self.windowOptions.title.enable then
		SendChatMessage(StripColorCodes(self.titleTextMethod(titleData)), chatType, nil, channel)
	end
	
	if limit then 
		if #sortList < limit then
			limit = #sortList
		end
	else
		limit = #sortList
	end

	for i = 1,limit do
		local text = Assessment:NewTable()
		local data = sortList[i]
		data.rank = i
		
		for j = 1,self.displaySet.text.count do
			text[j] = StripColorCodes( self.textMethods[j](data) )
		end
		
		SendChatMessage(table.concat(text, "  "), chatType, nil, channel)
		Assessment:ReleaseTable(text)
	end
end

-----------------------------------------------------------------------------
-----------------------------------------------------------------------------

function AssessmentBarInterface:CreateWindowOptions(i)
	local options = self:CopyOptionsTemplate(windowOptionsTemplate, i)
	options.name = L["Window %d"]:format(i)
	options.desc = L["Options for assessment window %d"]:format(i)
	return options
end

function AssessmentBarInterface:UpdateDisplaySetValidate()
	for k in pairs(displaySetValidate) do
		displaySetValidate[k] = nil
	end
	for k, v in AssessmentDisplaySets:IterateDisplaySets() do
		if v.selectable then
			displaySetValidate[k] = k
		end
	end
end

function AssessmentBarInterface:CreateOptions()
	self:ReleaseTable(Assessment.Options.args.windows.args, true)
	Assessment.Options.args.windows.args = self:CopyOptionsTemplate(windowsArgs)

	for i = 1, self.db.profile.windows.count do
		if Assessment.Options.args.windows.args[tostring(i)] then
			self:ReleaseTable(Assessment.Options.args.windows.args[tostring(i)], true)
		end
		Assessment.Options.args.windows.args[tostring(i)] = self:CreateWindowOptions(i)
	end
end

function AssessmentBarInterface:RemoveWindow(index)
	local windowOptions = self.db.profile.windows
	local windows = self.windows

	local oldWindow = windows[index]
	local oldWindowOptions = windowOptions[index]

	for i = index, windowOptions.count-1 do
		windowOptions[i] = windowOptions[i+1]
		windows[i] = windows[i+1]
	end
	windows[windowOptions.count] = oldWindow
	windowOptions[windowOptions.count] = oldWindowOptions
	windowOptions.count = windowOptions.count - 1
	
	oldWindow:Hide()	
	self:CreateOptions()
end

function AssessmentBarInterface:UpdateWindowOptions(index)
	self.windows[index]:SetWindowOptions(self.db.profile.windows[index])
end

function AssessmentBarInterface:OnInitialize()
	self.windows = { }

	if not self.db.profile.windows.count then
		self.db.profile.windows.count = 1
		self.db.profile.windows[1] = self:CopyTable(defaultWindowProfile)
	end
end

function AssessmentBarInterface:OnEnable()
	self:UpdateDisplaySetValidate()
	self:RegisterEvent("SharedMedia_SetGlobal", "UpdateAllWindows")
	self:RegisterEvent("AssessmentDisplaySets:OnDisplaySetAdded", "OnDisplaySetListChanged")
	self:RegisterEvent("AssessmentDisplaySets:OnDisplaySetRemoved", "OnDisplaySetListChanged")
	self:RegisterEvent("AssessmentDisplaySets:OnDisplaySetListChanged", "OnDisplaySetListChanged")
	self:RegisterEvent("AssessmentDisplaySets:OnDisplaySetRequestRemove", "OnDisplaySetRequestRemove")
	self:RegisterEvent("AssessmentDisplaySets:OnDisplaySetUpdated", "OnDisplaySetUpdated")
	self:CreateWindows()
	self:CreateOptions()
end

function AssessmentBarInterface:OnDisable()
end

function AssessmentBarInterface:UpdateAllWindows()
	for i = 1, self.db.profile.windows.count do
		local windowOptions = self.db.profile.windows[i]
		self.windows[i]:SetWindowOptions(windowOptions)
	end
end

function AssessmentBarInterface:CreateWindows()
	for i = 1, self.db.profile.windows.count do
		local windowOptions = self.db.profile.windows[i]
		if self.windows[i] then
			self.windows[i]:SetWindowOptions(windowOptions)
		else
			self.windows[i] = UIWindow:new(windowOptions)
		end
	end
	
	for i = self.db.profile.windows.count+1, #self.windows do
		self.windows[i]:Hide()
	end
end

function AssessmentBarInterface:GetWindowIndex(window)
	for i = 1,self.db.profile.windows.count do
		if self.windows[i] == window then
			return i
		end
	end
	error("INTERNAL ERROR: Window not found")
end

function AssessmentBarInterface:SetDisplaySet(window, displaySetName)
	Dewdrop:Close()
	local windowIndex = self:GetWindowIndex(window)
	local windowOptions = self.db.profile.windows[windowIndex]
	windowOptions.displaySetStackLevel = 1
	windowOptions.displaySetStack[1] = displaySetName
	self:UpdateWindowOptions(windowIndex)
end

function AssessmentBarInterface:PushWindowDisplaySet(window, displaySetName)
	local windowIndex = self:GetWindowIndex(window)
	local windowOptions = self.db.profile.windows[windowIndex]
	
	local displaySet = AssessmentDisplaySets:GetDisplaySet(("$"):split(displaySetName, 2))
	if not displaySet then
		error("Unable to set display set: "..displaySetName)
	end
	
	windowOptions.displaySetStackLevel = windowOptions.displaySetStackLevel + 1
	windowOptions.displaySetStack[windowOptions.displaySetStackLevel] = displaySetName
	self:UpdateWindowOptions(windowIndex)
end

function AssessmentBarInterface:PopWindowDisplaySet(window)
	local windowIndex = self:GetWindowIndex(window)
	local windowOptions = self.db.profile.windows[windowIndex]
	
	if windowOptions.displaySetStackLevel == 1 then return end
	
	windowOptions.displaySetStack[windowOptions.displaySetStackLevel] = nil
	windowOptions.displaySetStackLevel = windowOptions.displaySetStackLevel - 1
	self:UpdateWindowOptions(windowIndex)
end

function AssessmentBarInterface:GetDisplaySetName(windowOptions)
	return ("$"):split(windowOptions.displaySetStack[windowOptions.displaySetStackLevel], 2)
end

function AssessmentBarInterface:GetDisplaySet(windowOptions)
	local displaySetName, displaySetParameters = self:GetDisplaySetName(windowOptions)
	return AssessmentDisplaySets:GetDisplaySet(displaySetName), displaySetParameters
end

function AssessmentBarInterface:Report(windowIndex, chatType, limit, channel)
	self.windows[windowIndex]:Report(chatType, limit, channel)
end

function AssessmentBarInterface:AddWindow()
	local windowOptions = self.db.profile.windows
	windowOptions.count = windowOptions.count + 1
	
	if not windowOptions[windowOptions.count] then
		windowOptions[windowOptions.count] = self:CopyTable(defaultWindowProfile)
	end
	
	self:CreateWindows()
	self:CreateOptions()
end

-----------------------------------------------------------------------------

function AssessmentBarInterface:OnDisplaySetUpdated(displaySetName)
	for i = 1, self.db.profile.windows.count do
		local windowOptions = self.db.profile.windows[i]
		if ("$"):split(windowOptions.displaySetStack[windowOptions.displaySetStackLevel],2) == displaySetName then
			self.windows[i]:SetWindowOptions(windowOptions)
		end
	end
end

function AssessmentBarInterface:OnDisplaySetRequestRemove(displaySetName)
	for i = 1, self.db.profile.windows.count do
		local windowOptions = self.db.profile.windows[i]
		
		for j = 1, windowOptions.displaySetStackLevel do
			if ("$"):split(windowOptions.displaySetStack[j], 2) == displaySetName then
				AssessmentDisplaySets:PreventDisplaySetRemoval(L["This display set cannot be removed currently because one or more windows require its data."])
				return
			end
		end
	end
end

function AssessmentBarInterface:OnDisplaySetListChanged()
	self:UpdateDisplaySetValidate()
	self:CreateOptions()
	for i=1,self.db.profile.windows.count do
		local windowOptions = self.db.profile.windows[i]
		if self:GetDisplaySet(windowOptions).listType == "DisplaySetList" then
			self.windows[i].forceUpdate = true
		end
	end
end

-----------------------------------------------------------------------------
