-----------------------------------------------------------------------------

local AssessmentCombatTime = Assessment:NewModule("CombatTime", "AceEvent-2.0")
local Roster = AceLibrary("Roster-2.1")

-----------------------------------------------------------------------------

Assessment.TagData.combattime =
{
	method = function(data) 
				local fightData = data.fightData
				if not fightData then return 0 end
	
				local combatTime = fightData.combatTime 
				if not combatTime then return 0 end
				
				if combatTime.endTime then
					return combatTime.endTime - combatTime.startTime
				else
					return GetTime() - combatTime.startTime
				end
			 end,
	data = "combatTime",
	guaranteeResult = true,
	numeric = true
}

Assessment.TagData.isincombat =
{
	method = function(data, prior) 
				if not prior then return false end

				local fightData = data.fightData
				if not fightData then return false end
	
				local combatTime = fightData.combatTime 
				if not combatTime then return false end
				
				return not combatTime.endTime
			 end,
	data = "combatTime",
}
-----------------------------------------------------------------------------

function AssessmentCombatTime:OnInitialize()
end

function AssessmentCombatTime:OnEnable()
	self:RegisterEvent("Assessment:ResetFightData", "OnResetFightData")
	self:RegisterEvent("PLAYER_REGEN_DISABLED", "OnCombatStart")
	self:RegisterEvent("RosterLib_RosterUpdated")

	self:RosterLib_RosterUpdated()	
end

function AssessmentCombatTime:OnDisable()
end

-----------------------------------------------------------------------------

function AssessmentCombatTime:RosterLib_RosterUpdated()
	if GetNumPartyMembers() + GetNumRaidMembers() > 0 then
		self:ScheduleRepeatingEvent("Assessment:GroupCombatTimeUpdate", self.UpdateGroup, 0.2, self)
		self:CancelEvent("PLAYER_REGEN_ENABLED")
	else
		self:RegisterEvent("PLAYER_REGEN_ENABLED", "OnCombatEnd")
		self:CancelScheduledEvent("Assessment:GroupCombatTimeUpdate")
	end
end

-----------------------------------------------------------------------------

function AssessmentCombatTime:UpdateGroup()
	for unit in Roster:IterateRoster() do
		if UnitAffectingCombat(unit.unitid) then
			self:OnCombatStart()
			return
		end
	end
	self:OnCombatEnd()
end

-----------------------------------------------------------------------------

function AssessmentCombatTime:OnResetFightData()
	if inCombat then
		self:DoCombatStart()
	end
end

-----------------------------------------------------------------------------

local inCombat
function AssessmentCombatTime:OnCombatStart()
	if inCombat then return end
	inCombat = true
	
	self:DoCombatStart()
end

function AssessmentCombatTime:DoCombatStart()
	Assessment:StartNewFight()
	self:ScheduleRepeatingEvent("Assessment:CombatTimeUpdate", self.Update, 0, self)
	for data in self:IterateUpdateDataTables() do
		local combatTime = data.combatTime
		if combatTime then
			combatTime.endTime = nil
		else
			combatTime = self:NewTable()
			combatTime.startTime = GetTime()
			data.combatTime = combatTime
		end
	end
end

function AssessmentCombatTime:OnCombatEnd()
	if not inCombat then return end
	inCombat = nil

	self:CancelScheduledEvent("Assessment:CombatTimeUpdate")
	for data in self:IterateUpdateDataTables() do
		local combatTime = data.combatTime
		if not combatTime then
			combatTime = self:NewTable()
			combatTime.startTime = GetTime()
			data.combatTime = combatTime
		end
		combatTime.endTime = GetTime()
	end
end

function AssessmentCombatTime:Update()
	for data in self:IterateUpdateDataTables() do
		data.dataUpdateTimes.combatTime = GetTime()
	end
end

-----------------------------------------------------------------------------
