-----------------------------------------------------------------------------
--
-- data
--   \damageDealt
--     \<damageDealerName>
--       -count, amount, absorbAmount, blockAmount, resistAmount
--       \attackDetails
--         \<attackName>
--            -count, amount, absorbAmount, blockAmount, resistAmount
--            \hitDetails
--			    \<attackType> -- one of Hit, Critical Hit, Tick, Miss, Dodge, Resist, etc.
--                -count, amount, absorbAmount, blockAmount, resistAmount
--   \damageTaken
--     \<damageTakerName>
--       -count, amount, absorbAmount, blockAmount, resistAmount
--       \attackDetails
--         \<attackName>
--            -count, amount, absorbAmount, blockAmount, resistAmount
--            \hitDetails
--  			\<attackType>
--                -count, amount, absorbAmount, blockAmount, resistAmount
--
-----------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("AssessmentDamage")
L:RegisterTranslations("enUS", function() return {
	["Damage Dealers"] = true,
	["Damage Takers"] = true,
	["Damage Dealt"] = true,
	["Damage Dealt Details"] = true,
	["Damage Dealt Hit Details"] = true,
	["DPS"] = true,
	["Fight Damage Dealt Details"] = true,
	["Fight Damage Dealt Hit Details"] = true,
	["Hit"] = true,
	["Critical Hit"] = true,
	["Tick"] = true,
	["White Damage"] = true,
} end)

-----------------------------------------------------------------------------

local AssessmentDamage = Assessment:NewModule("Damage", "AceEvent-2.0", "Parser-3.0")
AssessmentDamage.db = Assessment:AcquireDBNamespace("Damage")
Assessment:RegisterDefaults("Damage", "profile", {
	enable = true
})

-----------------------------------------------------------------------------

Assessment.TagData.damagedealt =
{
	method = function(data)
				local fightData = data.fightData
				if not fightData then return 0 end
				
				local damageDealt = fightData.damageDealt
				if not damageDealt then return 0 end
				
				if not damageDealt[data.key] then return 0 end
				return damageDealt[data.key].amount
			 end,
	data = "damage",
	guaranteeResult = true,
	numeric = true,
}

Assessment.TagData.damagetaken =
{
	method = function(data)
				local fightData = data.fightData
				if not fightData then return 0 end
				
				local damageTaken = fightData.damageTaken
				if not damageTaken then return 0 end
				
				if not damageTaken[data.key] then return 0 end
				return damageTaken[data.key].amount
			 end,
	data = "damage",
	guaranteeResult = true,
	numeric = true,
}

Assessment.TagData.absorbamount =
{
	tableData = "value.absorbAmount",
	numeric = true
}

Assessment.TagData.blockamount =
{
	tableData = "value.blockAmount",
	numeric = true
}

Assessment.TagData.misscount =
{
	tableData = "value.count",
	numeric = true
}

Assessment.TagData.resistamount =
{
	tableData = "value.resistAmount",
	numeric = true
}

-----------------------------------------------------------------------------

Assessment.ListType.DamageDealers =
{
	method = function(fightData)
				if not fightData then return function() return nil end end
				return pairs(fightData.damageDealt)
			 end,
	data = "damage",
	description = L["Damage Dealers"],
}

Assessment.ListType.DamageTakers =
{
	method = function(fightData)
				if not fightData.damageTaken then return function() return nil end end
				return pairs(fightData.damageTaken)
			 end,
	data = "damage",
	description = L["Damage Takers"],
}

-----------------------------------------------------------------------------

Assessment.DefaultDisplaySets.DamageDealtHitDetails =
{
	titleTag			= "[windowDisplaySetParameter(2)]: [windowDisplaySetParameter(4)]",
	listType			= "EnumerateTable", 
	barValueTag			= "amount",
	filterTag			= "1",
	fightIndex			= 0,
	text				= { { tag = "[rank]. [key] ([count])" }, { tag = "[barValue:short]" } },
	description			= L["Damage Dealt Hit Details"],
	selectable			= false,
}

Assessment.DefaultDisplaySets.DamageDealtDetails =
{
	titleTag			= "Damage Dealt: [windowDisplaySetParameter(2)]",
	listType			= "EnumerateTable", 
	barValueTag			= "amount",
	filterTag			= "1",
	fightIndex			= 0,
	text				= { { tag = "[rank]. [key|\"White Damage\"] ([count])" }, { tag = "[barValue:short]" } },
	description			= L["Damage Dealt Details"],
	selectable			= false,
	onClickType			= "PUSH_DISPLAY_SET",
	onClickDisplaySet	= L["Damage Dealt Hit Details"],
	onClickData			= "damageDealt,[windowDisplaySetParameter(2)],attackDetails,[key],hitDetails"
}

Assessment.DefaultDisplaySets.DamageDealt =
{
	titleTag			= "Damage Dealt: [totalBarValue:short]",
	listType			= "GroupMembers", 
	barValueTag			= "damageDealt", 
	filterTag			= "barValue ~= 0",
	fightIndex			= 0,
	text				= { { }, { tag= "[barValue:short]" } },
	description			= L["Damage Dealt"],
	onClickType			= "PUSH_DISPLAY_SET",
	onClickDisplaySet	= L["Damage Dealt Details"],
	onClickData			= "damageDealt,[key],attackDetails"
}

Assessment.DefaultDisplaySets.FightDamageDealtHitDetails =
{
	titleTag			= "[windowDisplaySetParameter(2)]: [windowDisplaySetParameter(4)]",
	listType			= "EnumerateTable", 
	barValueTag			= "amount",
	filterTag			= "1",
	fightIndex			= 1,
	text				= { { tag = "[rank]. [key] ([count])" }, { tag = "[barValue:short]" } },
	description			= L["Fight Damage Dealt Hit Details"],
	selectable			= false,
}

Assessment.DefaultDisplaySets.FightDamageDealtDetails =
{
	titleTag			= "Fight Damage Dealt: [windowDisplaySetParameter(2)]",
	listType			= "EnumerateTable", 
	barValueTag			= "amount",
	filterTag			= "1",
	fightIndex			= 1,
	text				= { { tag = "[rank]. [key|\"White Damage\"] ([count])" }, { tag = "[barValue:short]" } },
	description			= L["Fight Damage Dealt Details"],
	selectable			= false,
	onClickType			= "PUSH_DISPLAY_SET",
	onClickDisplaySet	= L["Fight Damage Dealt Hit Details"],
	onClickData			= "damageDealt,[windowDisplaySetParameter(2)],attackDetails,[key],hitDetails"
}

Assessment.DefaultDisplaySets.DPS =
{
	titleTag			= "DPS: [combatTime:formatDuration]",
	listType			= "GroupMembers", 
	barValueTag			= "damageDealt / maximum(combatTime, 1)", 
	filterTag			= "barValue ~= 0",
	fightIndex			= 1,
	text				= { { }, { tag= "[barValue:format(\"%.1f\")]" } },
	description			= L["DPS"],
	onClickType			= "PUSH_DISPLAY_SET",
	onClickDisplaySet	= L["Fight Damage Dealt Details"],
	onClickData			= "damageDealt,[key],attackDetails"
}

-----------------------------------------------------------------------------

Assessment.DefaultTags.DPS =
{
	tag			= "[(damageDealt / maximum(combatTime, 1)):format(\"%.1f\")]",
	fightIndex	= 1,
	description	= L["DPS"],
}

-----------------------------------------------------------------------------

function AssessmentDamage:OnInitialize()
end

function AssessmentDamage:OnEnable()
	self:RegisterParserEvent({ eventType = "Damage" }, "OnParserDamageEvent")
end

function AssessmentDamage:OnDisable()
end

-----------------------------------------------------------------------------

local function AccumulateData(data, eventData)
	data.count = data.count + 1
	data.amount = data.amount + eventData.amount
	if eventData.absorbAmount ~= 0 then
		data.absorbAmount = (data.absorbAmount or 0) + eventData.absorbAmount
	end
	if eventData.blockAmount ~= 0 then
		data.blockAmount = (data.blockAmount or 0) + eventData.blockAmount
	end
	if eventData.resistAmount ~= 0 then
		data.resistAmount = (data.resistAmount or 0) + eventData.resistAmount
	end
end

-----------------------------------------------------------------------------

function AssessmentDamage:NewDataTable()
	local data = self:NewTable()
	data.count = 0
	data.amount = 0
	return data
end

-----------------------------------------------------------------------------

function AssessmentDamage:UpdateDamageData(dataTable, member, eventData, attackType, abilityName)
	local data = dataTable[member]
	if not data then
		data = self:NewDataTable()
		data.attackDetails = self:NewTable()
		dataTable[member] = data
	end
	AccumulateData(data, eventData)
	
	local attackDetailsParent = data.attackDetails
	local attackDetails = attackDetailsParent[abilityName]
	if not attackDetails then
		attackDetails = self:NewDataTable()
		attackDetails.hitDetails = self:NewTable()
		attackDetailsParent[abilityName] = attackDetails
	end
	AccumulateData(attackDetails, eventData)
	
	local hitDetails = attackDetails.hitDetails
	local attackTypeDetails = hitDetails[attackType]
	if not attackTypeDetails then
		attackTypeDetails = self:NewDataTable()
		hitDetails[attackType] = attackTypeDetails
	end
	AccumulateData(attackTypeDetails, eventData)
end

-----------------------------------------------------------------------------

function AssessmentDamage:OnParserDamageEvent(eventData)
	local attackType
	if eventData.isDoT then
		attackType = L["Tick"]
	elseif eventData.isCrit then
		attackType = L["Critical Hit"]
	else
		attackType = L["Hit"]
	end
	
	local abilityName = eventData.abilityName or L["White Damage"]
	
	for data in self:IterateUpdateDataTables() do
		data.dataUpdateTimes.damage = GetTime()
		local damageDealt = data.damageDealt
		if not damageDealt then
			damageDealt = self:NewTable()
			data.damageDealt = damageDealt
		end

		local damageTaken = data.damageTaken
		if not damageTaken then
			damageTaken = self:NewTable()
			data.damageTaken = damageTaken
		end

		self:UpdateDamageData(damageDealt, eventData.sourceName, eventData, attackType, abilityName)
		self:UpdateDamageData(damageTaken, eventData.recipientName, eventData, attackType, abilityName)
	end
end

-----------------------------------------------------------------------------
